"""parallel.py - Parallel Processing Utilities

This module provides utilities for performing parallel processing tasks using multithreading in Python.

Functions:
    - threaded_map: Multi-threaded mapping of a function to an iterable. Useful for I/O bound tasks.
"""

import threading
from time import sleep
from typing import Callable, Iterable

from tqdm.auto import tqdm


def threaded_map(
    target_function: Callable,
    args_list: Iterable,
    time_delay=0.02,
    max_n_threads=None,
    progress_bar=True,
):
    """
    Multi-threaded mapping of a function to an iterable. Useful for I/O bound tasks.

    This function allows you to apply a target function to elements in an iterable concurrently using multiple threads.
    You can control the number of threads, introduce time delays between thread launches, and enable a progress bar.

    Parameters:
        target_function (Callable): The function to apply to each element in the iterable.
        args_list (Iterable): The input data iterable to be processed in parallel.
        time_delay (float, optional): Time delay (in seconds) between launching threads. Default is 0.02.
        max_n_threads (int, optional): The maximum number of threads to run concurrently.
            Default is None, which means entire args_list will be processed concurrently.
        progress_bar (bool, optional): Enable or disable the progress bar. Default is True.

    Note:
        The progress bar provided by the tqdm library gives you visual feedback on the progress of thread launching and
        completion.

    Example:
        .. code-block:: python
            def get_webpage(url, results: dict):
                if not url in results:
                    results[url] = requests.get(url)
            urls = ["https://example.com", "https://example.com", ]
            results = {}
            args = [(url, results) for url in urls]
            threaded_map(get_webpage, args, max_n_threads=2)
            # Threads are launched concurrently, with a maximum of 2 threads at a time.
    """

    threads = []
    tqdm_bar = tqdm(args_list, desc="Launching threads") if progress_bar else None
    for n, args in enumerate(tqdm_bar or args_list):
        thread = threading.Thread(target=target_function, args=args, name=f"{n}_{args}")
        thread.start()
        threads.append(thread)
        sleep(time_delay)

        if max_n_threads:
            if len(threads) >= max_n_threads:
                for i, thread in enumerate(threads):
                    if tqdm_bar:
                        tqdm_bar.set_description(
                            f"Awaiting threads {i/len(threads):.0%}"
                        )
                    thread.join()
                threads = []
                if tqdm_bar:
                    tqdm_bar.set_description("Launching threads")

    for thread in tqdm(threads, desc="Awaiting threads"):
        thread.join()



"""This module provides utility functions for the sign language translator package.

Functions:
- download: A function for downloading files from urls.
- tree: A function for printing a directory tree.
- sample_one_index: Select an index based on the given probability distribution.
- search_in_values_to_retrieve_key: search inside every dict value and return the key on match.
- in_jupyter_notebook: Checks if the code is running in a Jupyter notebook.
- linear_interpolation: figure out intermediate values inside an array.
- threaded_map: Multi-threaded mapping of a function to an iterable.

Classes:
- ArrayOps: A class for array operations agnostic to numpy.ndarray and torch.Tensor.
"""

from sign_language_translator.utils.arrays import ArrayOps, linear_interpolation
from sign_language_translator.utils.download import download
from sign_language_translator.utils.parallel import threaded_map
from sign_language_translator.utils.tree import tree
from sign_language_translator.utils.utils import (
    in_jupyter_notebook,
    sample_one_index,
    search_in_values_to_retrieve_key,
)

__all__ = [
    # functions
    "tree",
    "download",
    "search_in_values_to_retrieve_key",
    "sample_one_index",
    "in_jupyter_notebook",
    "linear_interpolation",
    "threaded_map",
    # classes
    "ArrayOps",
]



from random import choices
from typing import Any, Dict, List, Set

__all__ = [
    "search_in_values_to_retrieve_key",
    "sample_one_index",
    "in_jupyter_notebook",
    "extract_recursive",
]


def search_in_values_to_retrieve_key(
    code_name: str, class_to_codes: Dict[Any, Set[str]]
):
    # verify there is no repetition/reuse in language codes
    all_codes = [code for codes in class_to_codes.values() for code in codes]
    assert len(all_codes) == len(set(all_codes)), "code reused for multiple keys"

    for key, codes in class_to_codes.items():
        if code_name.lower() in codes:
            return key

    return None


def sample_one_index(weights: List[float], temperature: float = 1.0) -> int:
    """Select an item based on the given probability distribution.
    Returns the index of the selected item sampled from weighted random distribution.

    Args:
        weights (List[float]): the relative weights corresponding to each index.
        temperature (float): The temperature value for controlling the sampling behavior.
            High temperature means sampling probabilities are more uniform (says random things).
            Low temperature means that sampling probabilities are higher for bigger weights.
            Defaults to 1.0.

    Returns:
        int: The index of the chosen item.
    """

    return choices(
        range(len(weights)),
        weights=[w / temperature for w in weights],
        k=1,
    )[0]


def in_jupyter_notebook():
    """
    Checks if the code is running in a Jupyter notebook.

    Returns:
        bool: True if running in a Jupyter notebook, False otherwise.
    """

    try:
        from IPython import get_ipython  # type: ignore

        return "IPKernelApp" in get_ipython().config  # type: ignore
    except:  # pylint: disable = bare-except
        return False


def extract_recursive(data: Dict[str, Any], key: str) -> List[Any]:
    """
    Recursively extracts values associated with a specified key from a nested dictionary.

    Args:
        data (Dict[str, Any]): The input dictionary containing nested structures.
        key (str): The key for which values need to be extracted.

    Returns:
        List[Any]: A list containing all values associated with the specified key, extracted
                   recursively from the input dictionary.

    Examples:
        >>> data = {'a': 1, 'b': {'c': 2, 'd': {'e': 3, 'f': 4}}, 'g': [5, {'h': 6, 'e': 7}]}
        >>> extract_recursive(data, 'e')
        [3, 7]
        >>> extract_recursive(data, 'h')
        [6]
        >>> extract_recursive(data, 'x')
        []  # Key not found, returns an empty list.
    """

    extracted_values = []

    def extract(data: Dict, results: List):
        for k in data:
            if k == key:
                results.append(data[k])
            elif isinstance(data[k], dict):
                extract(data[k], results)
            elif isinstance(data[k], list):
                for item in data[k]:
                    if isinstance(item, dict):
                        extract(item, results)

    extract(data, extracted_values)

    return extracted_values



"""Module for managing assets required by to the sign language translator,
such as datasets and models."""

import json
import re
from datetime import datetime
from os import makedirs, remove
from os.path import abspath, dirname, exists, isdir, isfile, join, sep
from typing import Dict, List, Tuple

from sign_language_translator.config.settings import Settings
from sign_language_translator.config.utils import read_urls
from sign_language_translator.utils.download import download


class Assets:
    """
    Class for managing assets related to the sign language translator.

    Attributes:
        ROOT_DIR (str): The root directory path where the sign language datasets & models are stored.
        FILE_TO_URL (Dict[str, str]): A dictionary mapping asset filenames to their corresponding URLs.
        primary_urls_file (str): The name of the first URLs file that is loaded by default and must contain links to other url files.
        urls_file_dir (str): The directory path where the URLs files are stored.
        asset_regex_to_urls_file (List[Tuple[str, str]]): A list of tuples mapping regular expressions matching asset names to URLs files.
        _checksum_filename (str): The filename for the checksum file.
        _loaded_url_files (set): A set containing names of loaded URLs files.

    Methods:
        set_root_dir(path: str) -> None:
            Set the SLT resources directory path.

        get_id(filename_or_regex: str) -> List[str]:
            Get the IDs matching the given filename or regex.

        get_url(filename_or_regex: str) -> List[str]:
            Get the URLs for assets matching the given filename or regex.

        get_path(filename_or_regex: str) -> List[str]:
            Get the file paths for assets matching the given filename or regex.

        download(filename_or_regex: str, overwrite=False, progress_bar: bool | None = None, timeout: float = 20.0, leave=True, chunk_size=65536) -> bool:
            Download assets matching the given filename regex and save them to the appropriate file paths inside the assets root directory.

        delete_out_of_date_asset(asset_id: str, url: str, checksum: Dict[str, Dict[str, str]]) -> None:
            Delete assets that are out of date based on the checksum (change in URL).

        reload() -> None:
            Clear cache and read the URL files again.

        load_urls(filename: str) -> None:
            Load URLs from the specified file.
    """

    # =============== #
    #    CONFIGURE    #
    # =============== #

    ROOT_DIR: str = join(dirname(dirname(abspath(__file__))), "assets")
    """The root directory path where the sign language datasets & models are stored. Defaults is 'install_directory/assets'."""

    primary_urls_file: str = "urls.json"
    """The name of the first URLs file that is loaded by default and must contain links to other url files.
    Defaults is 'urls.json'. Note: all url filenames must end with 'urls.json'."""

    urls_file_dir: str = dirname(abspath(__file__))
    """The directory path where the URLs files are stored. Defaults is 'install_directory/sign_language_translator/config'."""

    FILE_TO_URL: Dict[str, str] = read_urls(join(urls_file_dir, primary_urls_file))
    """A dictionary mapping asset filenames to their corresponding URLs."""

    asset_regex_to_urls_file: List[Tuple[str, str]] = [
        # regex_for_assets, urls_file
        # (r"videos/pk-.*\.mp4, "pk-reference-clips-urls.json")
        (r".*", "extra_urls.json"),
    ]
    """A list of tuples that map regular expressions matching asset names to URLs files containing URLs to that asset group.
    The regex are tried sequentially so make sure list order is right."""

    _checksum_filename = ".checksum.json"
    _loaded_url_files = set()

    @classmethod
    def set_root_dir(cls, path: str) -> None:
        """Set the SLT resources directory path.
        Helpful when using custom datasets, or when data must be stored outside install directory.
        (e.g hosted on cloud and mounted on disk).
        By default, resources are downloaded to the 'install_directory/assets'.

        Args:
            path (str): The path to the resources/dataset/models directory.

        Raises:
            ValueError: If the provided path is not a directory.
        """

        path = abspath(path)
        if exists(path) and not isdir(path):
            raise ValueError(f"The provided path is not a directory. Path: {path}")
        makedirs(path, exist_ok=True)

        cls.ROOT_DIR = path

    # ============= #
    #    Getters    #
    # ============= #

    @classmethod
    def get_id(cls, filename_or_regex: str) -> List[str]:
        """
        Get the list of asset IDs that match the given filename or regex.

        Args:
            filename_or_regex (str): The filename or regex to match against asset IDs.

        Returns:
            List[str]: List of matching asset IDs.
        """

        # exact match
        if filename_or_regex in cls.FILE_TO_URL:
            return [filename_or_regex]

        # regex match
        try:
            regex = re.compile(filename_or_regex)
            ids = [file for file in cls.FILE_TO_URL if re.match(regex, file)]
            if ids:
                return ids
        except re.error:
            pass

        # load more urls
        for regex, urls_file in cls.asset_regex_to_urls_file:
            if urls_file in cls._loaded_url_files:
                continue

            if regex == filename_or_regex or re.match(regex, filename_or_regex):
                cls.load_urls(urls_file)

                # recursive call
                return cls.get_id(filename_or_regex)

        # not found
        return []

    @classmethod
    def get_url(cls, filename_or_regex: str) -> List[str]:
        """
        Get the list of URLs corresponding to the given filename or regex.

        Args:
            filename_or_regex (str): The filename or regex to match against asset IDs.

        Returns:
            List[str]: List of matching URLs.
        """

        return [cls.FILE_TO_URL[id] for id in cls.get_id(filename_or_regex)]

    @classmethod
    def get_path(cls, filename_or_regex: str) -> List[str]:
        """
        Get the list of file paths corresponding to the given filename or regex.

        Args:
            filename_or_regex (str): The filename or regex to match against asset IDs.

        Returns:
            List[str]: List of matching file paths.
        """

        return [
            join(
                (cls.urls_file_dir if asset_id.endswith("urls.json") else cls.ROOT_DIR),
                asset_id.replace("/", sep),
            )
            for asset_id in cls.get_id(filename_or_regex)
        ]

    # ================ #
    #    Management    #
    # ================ #

    @classmethod
    def download(
        cls,
        filename_or_regex: str,
        overwrite=False,
        progress_bar: bool | None = None,
        timeout: float = 20.0,
        leave=True,
        chunk_size=65536,
    ) -> bool:
        """
        Downloads package assets matching the given filename regex and saves them to the appropriate file paths.

        Args:
            filename_or_regex (str): Relative path or Regular expression to match the desired asset names.
            overwrite (bool, optional): If False, skips downloading if the resource file already exists. Defaults to False.
            progress_bar (bool, optional): If True, displays a progress bar during the download. If None, uses the value in slt.Settings.SHOW_DOWNLOAD_PROGRESS. Defaults to None.
            timeout (float, optional): The maximum number of seconds to wait for a server response. Defaults to 20.0.
            leave (bool, optional): Wether to leave the progress bar behind after the download. Defaults to True.
            chunk_size (int, optional): The number of bytes to fetch in each step. Defaults to 65536.

        Returns:
            bool: True if all resources are downloaded successfully or already exist, False otherwise.
        """

        if progress_bar is None:
            progress_bar = Settings.SHOW_DOWNLOAD_PROGRESS
            leave = False

        matching_ids = cls.get_id(filename_or_regex)
        urls = [cls.get_url(m_id)[0] for m_id in matching_ids]

        checksum = cls._read_checksum()

        statuses = []
        for asset_id, url in zip(matching_ids, urls):
            cls.delete_out_of_date_asset(asset_id, url, checksum)

            # Make sure that the file/directory exists
            file_path = cls.get_path(asset_id)[0]
            if exists(file_path) and not overwrite:
                continue
            makedirs(dirname(file_path), exist_ok=True)

            # Download the file from the URL
            status = download(
                file_path,
                url,
                progress_bar=progress_bar,
                timeout=timeout,
                overwrite=overwrite,
                leave=leave,
                chunk_size=chunk_size,
            )
            statuses.append(status)

            # update checksum file with date, url, filename, n_bytes
            if status:
                checksum[asset_id] = {
                    "date": datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "url": url,
                }
                cls._save_checksum(checksum)

        return all(statuses or [False])

    @classmethod
    def delete_out_of_date_asset(
        cls, asset_id: str, url: str, checksum: Dict[str, Dict[str, str]]
    ) -> None:
        """
        Delete an asset if it is out of date. Determined by comparing the URL the asset
        was downloaded from differs from with the loaded URL.

        Args:
            asset_id (str): The ID of the asset to check.
            url (str): The URL associated with the asset.
            checksum (Dict[str, Dict[str, str]]): The checksum information for assets.
        """

        if (
            exists((path := cls.get_path(asset_id)[0]))
            and checksum.get(asset_id, {}).get("url", None) != url
        ):
            remove(path)
            checksum.pop(asset_id, None)

    # TODO: delete all out of date assets on import
    # @classmethod
    # def delete_all_out_of_date_assets(cls):
    #     """
    #     Delete all out of date assets determined by the checksum file and current URLs.
    #     """
    #     cls.reload()
    #     checksum = cls._read_checksum()
    #     downloaded_assets = list(checksum.keys())
    #     for asset_id in downloaded_assets:
    #         cls.delete_out_of_date_asset(asset_id, cls.get_url(asset_id)[0], checksum)
    #     cls._save_checksum(checksum) # bug: not saving checksum for urls file but downloads it

    @classmethod
    def reload(cls) -> None:
        """
        Clear the cache and read the URL files again.
        """

        cls.FILE_TO_URL = read_urls(join(cls.urls_file_dir, cls.primary_urls_file))
        cls._loaded_url_files = set()

    @classmethod
    def load_urls(cls, filename: str) -> None:
        """
        Load URLs from the specified file into the Assets.FILE_TO_URL dictionary.

        Args:
            filename (str): The name of the URLs file to load.
        """
        cls.download(filename)
        file_ids = cls.get_id(filename)
        urls_file_paths = [cls.get_path(f_id)[0] for f_id in file_ids]
        for file_id, path in zip(file_ids, urls_file_paths):
            cls.FILE_TO_URL.update(read_urls(path))
            cls._loaded_url_files.add(file_id)

    @classmethod
    def _read_checksum(cls) -> Dict[str, Dict[str, str]]:
        makedirs(cls.ROOT_DIR, exist_ok=True)
        checksum_file_path = join(cls.ROOT_DIR, cls._checksum_filename)
        if isfile(checksum_file_path):
            with open(checksum_file_path, "r", encoding="utf-8") as f:
                checksum: Dict[str, Dict[str, str]] = json.load(f)
        else:
            checksum = {}
            cls._save_checksum(checksum)

        return checksum

    @classmethod
    def _save_checksum(cls, checksum: Dict[str, Dict[str, str]]) -> None:
        checksum_file_path = join(cls.ROOT_DIR, cls._checksum_filename)
        with open(checksum_file_path, "w", encoding="utf-8") as f:
            json.dump(checksum, f, indent=2, ensure_ascii=False, sort_keys=True)



"""
sign_language_translator
========================
Code: https://github.com/sign-language-translator/sign-language-translator
Help: https://sign-language-translator.readthedocs.io

This project is an effort to bridge the communication gap between the hearing and the hearing-impaired community using Artificial Intelligence.
The goal is to provide a user friendly API to novel Sign Language Translation solutions that can easily adapt to any regional sign language.

Usage
-----

.. code-block:: python

    import sign_language_translator as slt

    # download dataset or models (if you need them for personal use)
    # (by default, resources are auto-downloaded within the install directory)
    # slt.Assets.set_root_dir("path/to/folder")  # Helps preventing duplication across environments or using cloud synced data
    # slt.Assets.download(".*.json")  # downloads into resource_dir
    # print(slt.Settings.FILE_TO_URL.keys())  # All downloadable resources

    print("All available models:")
    print(list(slt.ModelCodes))  # slt.ModelCodeGroups
    # print(list(slt.TextLanguageCodes))
    # print(list(slt.SignLanguageCodes))
    # print(list(slt.SignFormatCodes))

    # -------------------------- TRANSLATE: text to sign --------------------------

    import sign_language_translator as slt

    # Load text-to-sign model
    # deep_t2s_model = slt.get_model("t2s-flan-T5-base-01.pt") # pytorch

    # rule-based model (concatenates clips of each word)
    t2s_model = slt.models.ConcatenativeSynthesis(
        text_language = "urdu", # or object of any child of slt.languages.text.text_language.TextLanguage class
        sign_language = "pakistan-sign-language", # or object of any child of slt.languages.sign.sign_language.SignLanguage class
        sign_format = "video", # or object of any child of slt.vision.sign.Sign class
    )

    text = "HELLO دنیا!" # HELLO treated as an acronym
    sign_language_sentence = t2s_model(text)

    # sign_language_sentence.show() # class: slt.vision.sign.Sign or its child
    # sign_language_sentence.save(f"sentences/{text}.mp4")

    # -------------------------- TRANSLATE: sign to text --------------------------

    import sign_language_translator as slt

    # load sign
    video = slt.Video("video.mp4")
    # features = slt.extract_features(video, "mediapipe_pose_v2_hand_v1")

    # Load sign-to-text model
    # deep_s2t_model = slt.get_model(slt.ModelCodes.TRANSFORMER_S2T) # pytorch
    deep_s2t_model = slt.get_model("gesture_mp_base-01") # pytorch

    # translate via single call to pipeline
    # text = deep_s2t_model.translate(video)

    # translate via individual steps
    features = deep_s2t_model.extract_features(video.iter_frames())
    encoding = deep_s2t_model.encoder(features)
    # logits = deep_s2t_model.decoder(encoding, token_ids = [0])
    # logits = deep_s2t_model.decoder(encoding, token_ids = [0, logits.argmax(dim=-1)])
    # ...
    tokens = deep_s2t_model.decode(encoding) # uses beam search to generate a token sequence
    text = "".join(tokens) # deep_s2t_model.detokenize(tokens)

    print(features.shape)
    print(logits.shape)
    print(text)
"""

from sign_language_translator import (
    config,
    data_collection,
    languages,
    models,
    text,
    utils,
    vision,
)
from sign_language_translator.config import enums
from sign_language_translator.config.assets import Assets
from sign_language_translator.config.enums import ModelCodeGroups, ModelCodes
from sign_language_translator.config.enums import SignFormats as SignFormatCodes
from sign_language_translator.config.enums import SignLanguages as SignLanguageCodes
from sign_language_translator.config.enums import TextLanguages as TextLanguageCodes
from sign_language_translator.config.settings import Settings
from sign_language_translator.languages import get_sign_language, get_text_language
from sign_language_translator.models import get_model
from sign_language_translator.vision._utils import get_sign_wrapper_class
from sign_language_translator.vision.video.video import Video

__version__ = config.utils.get_package_version()

__all__ = [
    # config
    "Assets",
    "Settings",
    "__version__",
    # modules
    "vision",
    "text",
    "data_collection",
    "models",
    "languages",
    "utils",
    "config",
    "enums",
    # classes (enum)
    "ModelCodes",
    "TextLanguageCodes",
    "SignLanguageCodes",
    "SignFormatCodes",
    "ModelCodeGroups",
    # classes (wrappers)
    "Video",
    # object loaders
    "get_sign_language",
    "get_text_language",
    "get_model",
    "get_sign_wrapper_class",
]

# TODO: Assets.delete_all_out_of_date_assets()
